@@ -1,22 +1,26 @@
 use std::{fmt::Display, iter::once};
 
 use crate::{
-    combi,
+    before, combi,
     error::{Builder as Eb, CustomBuilder as Cb},
-    prim::Error,
+    not_followed_by,
     util::run_drop,
     ICont, IOk, IReturn, Input, Parser, ParserOnce,
 };
 
 impl<'a, I: Input, C, S, M: Cb> ICont<'a, I, C, S, M> {
     #[inline]
-    pub fn pure<O>(self, value: O) -> IReturn<'a, O, I, C, S, M> {
+    pub fn to<O>(self, value: O) -> IReturn<'a, O, I, C, S, M> {
         IReturn(Ok((value, self)))
     }
     #[inline]
-    pub fn pure_or<P: Parser<I, C, S, M>>(self, o: Option<P::Output>, p: P) -> IReturn<'a, P::Output, I, C, S, M> {
+    pub fn done(self) -> IReturn<'a, (), I, C, S, M> {
+        IReturn(Ok(((), self)))
+    }
+    #[inline]
+    pub fn to_or<P: Parser<I, C, S, M>>(self, o: Option<P::Output>, p: P) -> IReturn<'a, P::Output, I, C, S, M> {
         match o {
-            Some(o) => self.pure(o),
+            Some(o) => self.to(o),
             None => self.then(p),
         }
     }
@@ -26,19 +30,16 @@ impl<'a, I: Input, C, S, M: Cb> ICont<'a, I, C, S, M> {
         IReturn(p.run_once(ICont { config, drop, ok }).map(|(o, ok)| (o, ok.to_cont(config, drop))))
     }
     #[inline]
-    pub fn fail<O, E: Display + 'static>(self, err: Error<E>) -> IReturn<'a, O, I, C, S, M> {
-        IReturn(Err(match err {
-            Error::Unexpect(token) => Eb::unexpected(token),
-            Error::Message(msg) => Eb::message(msg),
-        }
-        .at::<I>(self.ok.input.index(), self.ok.input.pos(), None)))
+    pub fn fail<O>(self, err: Eb<M>) -> IReturn<'a, O, I, C, S, M> {
+        IReturn(Err(err.at::<I>(self.ok.input.index(), self.ok.input.pos(), None)))
     }
     #[inline]
     pub fn config<O, F: FnOnce(&C, Self) -> IReturn<'a, O, I, C, S, M>>(self, f: F) -> IReturn<'a, O, I, C, S, M> {
         f(self.config, self)
     }
-
-    pub fn tail_rec<O1, O2>(self, o1: O1, f: impl Fn(O1, Self) -> IReturn<'a, Result<O2, O1>, I, C, S, M>) -> IReturn<'a, O2, I, C, S, M> {
+    pub fn tail_rec<O1, O2>(
+        self, o1: O1, f: impl Fn(O1, Self) -> IReturn<'a, Result<O2, O1>, I, C, S, M>,
+    ) -> IReturn<'a, O2, I, C, S, M> {
         match f(o1, self).0 {
             Err(e) => IReturn(Err(e)),
             Ok((Err(o), ok)) => ok.tail_rec(o, f),
@@ -46,18 +47,34 @@ impl<'a, I: Input, C, S, M: Cb> ICont<'a, I, C, S, M> {
         }
     }
 }
-impl<'a, I: Input + Clone, C, S: Clone, M: Cb> ICont<'a, I, C, S, M> {
-    pub fn fold<O, P: Parser<I, C, S, M>>(self, o1: O, p: P, f: impl Fn(O, P::Output) -> O) -> IReturn<'a, O, I, C, S, M> {
+impl<'a, I: Input, C, S: Clone, M: Cb> ICont<'a, I, C, S, M> {
+    #[inline]
+    pub fn before<P: ParserOnce<I, C, S, M>>(self, p: P) -> IReturn<'a, P::Output, I, C, S, M> {
+        self.then(before(p))
+    }
+    #[inline]
+    pub fn not_followed_by(
+        self, p: impl ParserOnce<I, C, S, M>, label: impl Display + 'static,
+    ) -> IReturn<'a, (), I, C, S, M> {
+        self.then(not_followed_by(p, label))
+    }
+    pub fn fold<O, P: Parser<I, C, S, M>>(
+        self, o1: O, p: P, f: impl Fn(O, P::Output) -> O,
+    ) -> IReturn<'a, O, I, C, S, M> {
         let ICont { ok, config, drop } = self;
         let (input2, state2, cutted) = (ok.input.clone(), ok.state.clone(), ok.cutted);
         match run_drop(p.to_ref(), ICont { ok, config, drop }, (input2, state2)) {
             (Err(e), None) => IReturn(Err(e)),
-            (Err(e), Some((input, state))) => IReturn(Ok((o1, IOk { input, state, err: Some(e), cutted }.to_cont(config, drop)))),
+            (Err(e), Some((input, state))) => {
+                IReturn(Ok((o1, IOk { input, state, err: Some(e), cutted }.to_cont(config, drop))))
+            },
             (Ok((o2, ok)), _) => ok.to_cont(config, drop).fold(f(o1, o2), p, f),
         }
     }
     #[inline]
-    pub fn fold1<O, P: Parser<I, C, S, M>>(self, o1: O, p: P, f: impl Fn(O, P::Output) -> O) -> IReturn<'a, O, I, C, S, M> {
+    pub fn fold1<O, P: Parser<I, C, S, M>>(
+        self, o1: O, p: P, f: impl Fn(O, P::Output) -> O,
+    ) -> IReturn<'a, O, I, C, S, M> {
         self.then(p.to_ref()).case(|o2, k| k.fold(f(o1, o2), p, f))
     }
 
@@ -83,13 +100,6 @@ impl<'a, I: Input + Clone, C, S: Clone, M: Cb> ICont<'a, I, C, S, M> {
         self.then(combi::SepFold { init: o, p, sep, succ: f })
     }
 
-    // #[inline]
-    // pub fn sep1<O, P1: Parser<I, C, S, M>, P2: Parser<I, C, S, M>>(
-    //     self, o: O, p: P1, sep: P2, f: impl Fn(O, P1::Output) -> O,
-    // ) -> IReturn<'a, O, I, C, S, M> {
-    //     self.then(combi::Sep1 { init: o, p, sep, succ: f })
-    // }
-
     #[inline]
     pub fn sep_extend<O: Extend<P1::Output>, P1: Parser<I, C, S, M>, P2: Parser<I, C, S, M>>(
         self, o: O, p: P1, sep: P2,
@@ -105,7 +115,9 @@ impl<'a, I: Input + Clone, C, S: Clone, M: Cb> ICont<'a, I, C, S, M> {
 }
 impl<'a, O, I: Input, C, S, M: Cb> IReturn<'a, O, I, C, S, M> {
     #[inline]
-    pub fn case<O2>(self, f: impl FnOnce(O, ICont<'a, I, C, S, M>) -> IReturn<'a, O2, I, C, S, M>) -> IReturn<'a, O2, I, C, S, M> {
+    pub fn case<O2>(
+        self, f: impl FnOnce(O, ICont<'a, I, C, S, M>) -> IReturn<'a, O2, I, C, S, M>,
+    ) -> IReturn<'a, O2, I, C, S, M> {
         IReturn(self.0.and_then(|(o, k)| f(o, k).0))
     }
     #[inline]
@@ -113,10 +125,14 @@ impl<'a, O, I: Input, C, S, M: Cb> IReturn<'a, O, I, C, S, M> {
         IReturn(self.0.map(|(o, k)| (f(o), k)))
     }
     #[inline]
-    pub fn value<O2>(self, o: O2) -> IReturn<'a, O2, I, C, S, M> {
+    pub fn to<O2>(self, o: O2) -> IReturn<'a, O2, I, C, S, M> {
         IReturn(self.0.map(|(_, k)| (o, k)))
     }
     #[inline]
+    pub fn done(self) -> IReturn<'a, (), I, C, S, M> {
+        IReturn(self.0.map(|(_, k)| ((), k)))
+    }
+    #[inline]
     pub fn bind<P: ParserOnce<I, C, S, M>>(self, f: impl FnOnce(O) -> P) -> IReturn<'a, P::Output, I, C, S, M> {
         self.case(|o, k| k.then(f(o)))
     }
@@ -126,14 +142,16 @@ impl<'a, O, I: Input, C, S, M: Cb> IReturn<'a, O, I, C, S, M> {
     }
     #[inline]
     pub fn left<P: ParserOnce<I, C, S, M>>(self, p: P) -> IReturn<'a, O, I, C, S, M> {
-        self.case(|o, k| k.then(p).value(o))
+        self.case(|o, k| k.then(p).to(o))
     }
     #[inline]
     pub fn right<P: ParserOnce<I, C, S, M>>(self, p: P) -> IReturn<'a, P::Output, I, C, S, M> {
         self.case(|_, k| k.then(p))
     }
     #[inline]
-    pub fn tail_rec<O2>(self, f: impl Fn(O, ICont<I, C, S, M>) -> IReturn<'a, Result<O2, O>, I, C, S, M>) -> IReturn<'a, O2, I, C, S, M> {
+    pub fn tail_rec<O2>(
+        self, f: impl Fn(O, ICont<I, C, S, M>) -> IReturn<'a, Result<O2, O>, I, C, S, M>,
+    ) -> IReturn<'a, O2, I, C, S, M> {
         self.case(|o, k| k.tail_rec(o, f))
     }
     #[inline]
@@ -142,7 +160,7 @@ impl<'a, O, I: Input, C, S, M: Cb> IReturn<'a, O, I, C, S, M> {
     }
 }
 
-impl<'a, O, I: Input + Clone + 'a, C, S: Clone + 'a, M: Cb> IReturn<'a, O, I, C, S, M> {
+impl<'a, O, I: Input, C, S: Clone, M: Cb> IReturn<'a, O, I, C, S, M> {
     #[inline]
     pub fn fold<P: Parser<I, C, S, M>>(self, p: P, f: impl Fn(O, P::Output) -> O) -> IReturn<'a, O, I, C, S, M> {
         self.case(|o, k| k.fold(o, p, f))
@@ -165,7 +183,10 @@ impl<'a, O, I: Input + Clone + 'a, C, S: Clone + 'a, M: Cb> IReturn<'a, O, I, C,
     {
         self.case(|o, k| k.extend1(o, p))
     }
-
+    #[inline]
+    pub fn not_followed_by(self, p: impl ParserOnce<I, C, S, M>, label: impl Display + 'static) -> Self {
+        self.left(not_followed_by(p, label))
+    }
     #[inline]
     pub fn sep_fold<P1: Parser<I, C, S, M>, P2: Parser<I, C, S, M>>(
         self, p: P1, sep: P2, f: impl Fn(O, P1::Output) -> O,
@@ -173,20 +194,19 @@ impl<'a, O, I: Input + Clone + 'a, C, S: Clone + 'a, M: Cb> IReturn<'a, O, I, C,
         self.case(|o, k| k.sep_fold(o, p, sep, f))
     }
 
-    // #[inline]
-    // pub fn sep1<P1: Parser<I, C, S, M>, P2: Parser<I, C, S, M>>(self, p: P1, sep: P2, f: impl Fn(O, P1::Output) -> O) -> IReturn<'a, O, I, C, S, M> {
-    //     self.case(|o, k| k.sep1(o, p, sep, f))
-    // }
-
     #[inline]
-    pub fn sep_extend<P1: Parser<I, C, S, M>, P2: Parser<I, C, S, M>>(self, p: P1, sep: P2) -> IReturn<'a, O, I, C, S, M>
+    pub fn sep_extend<P1: Parser<I, C, S, M>, P2: Parser<I, C, S, M>>(
+        self, p: P1, sep: P2,
+    ) -> IReturn<'a, O, I, C, S, M>
     where
         O: Extend<P1::Output>,
     {
         self.case(|o, k| k.sep_extend(o, p, sep))
     }
     #[inline]
-    pub fn sep_extend1<P1: Parser<I, C, S, M>, P2: Parser<I, C, S, M>>(self, p: P1, sep: P2) -> IReturn<'a, O, I, C, S, M>
+    pub fn sep_extend1<P1: Parser<I, C, S, M>, P2: Parser<I, C, S, M>>(
+        self, p: P1, sep: P2,
+    ) -> IReturn<'a, O, I, C, S, M>
     where
         O: Extend<P1::Output>,
     {
