@@ -1,9 +1,11 @@
-use std::{fmt::Display, marker::PhantomData};
+use std::{fmt::Display, marker::PhantomData, ops::RangeBounds};
 
 use either::Either;
 
 use crate::{
-    error::{Builder, Builder as Eb, CustomBuilder as Cb},
+    error::{Builder as Eb, CustomBuilder as Cb},
+    input::IntoChars,
+    util::run_satisfy,
     ICont, IOk, IResult, IReturn, Input, Parser, ParserOnce,
 };
 
@@ -29,15 +31,15 @@ impl<I: Input, C, S, M: Cb, P: ParserOnce<I, C, S, M>, F: Fn() -> P> Parser<I, C
 ///     // Note the reverse order
 ///     parser_once(|k| k.then(any.and(p).map(|(x,mut xs)| {xs.push(x); xs}).or(pure(String::new()))))
 /// }
-/// assert_eq!(p.parse_ok("abcd".chars()), Some("dcba".to_string()));
-/// assert_eq!(p.parse_ok("stressed".chars()), Some("desserts".to_string()));
+/// assert_eq!(p.parse_ok("abcd"), Some("dcba".to_string()));
+/// assert_eq!(p.parse_ok("stressed"), Some("desserts".to_string()));
 ///
 /// // Straightforward implementation without wasting stacks
 /// let q = tail_rec(vec![], |mut xs| any.or_not().map_once(move |x| match x {
 ///     Some(x) => { xs.push(x); Err(xs) },
 ///     None => Ok(xs)
 /// })).map(|xs| xs.iter().rev().collect::<String>());
-/// assert_eq!(q.parse_ok("stressed".chars()), Some("desserts".to_string()));
+/// assert_eq!(q.parse_ok("stressed"), Some("desserts".to_string()));
 /// ```
 #[derive(Clone, Copy)]
 pub struct FnParser<F>(F);
@@ -46,10 +48,14 @@ pub fn parser<I: Input, O, C, S, M: Cb, F: Fn(ICont<I, C, S, M>) -> IReturn<O, I
     FnParser(f)
 }
 #[inline]
-pub fn parser_once<I: Input, O, C, S, M: Cb, F: FnOnce(ICont<I, C, S, M>) -> IReturn<O, I, C, S, M>>(f: F) -> FnParser<F> {
+pub fn parser_once<I: Input, O, C, S, M: Cb, F: FnOnce(ICont<I, C, S, M>) -> IReturn<O, I, C, S, M>>(
+    f: F,
+) -> FnParser<F> {
     FnParser(f)
 }
-impl<O, I: Input, C, S, M: Cb, F: FnOnce(ICont<I, C, S, M>) -> IReturn<O, I, C, S, M>> ParserOnce<I, C, S, M> for FnParser<F> {
+impl<O, I: Input, C, S, M: Cb, F: FnOnce(ICont<I, C, S, M>) -> IReturn<O, I, C, S, M>> ParserOnce<I, C, S, M>
+    for FnParser<F>
+{
     type Output = O;
     #[inline]
     fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<O, I, S, M> {
@@ -68,8 +74,8 @@ impl<O, I: Input, C, S, M: Cb, F: Fn(ICont<I, C, S, M>) -> IReturn<O, I, C, S, M
 /// use chasa::*;
 /// let p = char('a').and(char('b'));
 /// let p = p.to_ref();
-/// assert_eq!(p.parse_ok("ab".chars()), Some(('a','b')));
-/// assert_eq!(many(p).parse_ok("abababc".chars()), Some(vec![('a','b'),('a','b'),('a','b')]));
+/// assert_eq!(p.parse_ok("ab"), Some(('a','b')));
+/// assert_eq!(many(p).parse_ok("abababc"), Some(vec![('a','b'),('a','b'),('a','b')]));
 /// ```
 pub struct RefParser<'a, P>(pub(crate) &'a P);
 impl<'a, P> Clone for RefParser<'a, P> {
@@ -93,7 +99,9 @@ impl<'a, I: Input, C, S, M: Cb, P: Parser<I, C, S, M>> Parser<I, C, S, M> for Re
     }
 }
 
-impl<I: Input, C, S, M: Cb, P1: ParserOnce<I, C, S, M>, P2: ParserOnce<I, C, S, M, Output = P1::Output>> ParserOnce<I, C, S, M> for Either<P1, P2> {
+impl<I: Input, C, S, M: Cb, P1: ParserOnce<I, C, S, M>, P2: ParserOnce<I, C, S, M, Output = P1::Output>>
+    ParserOnce<I, C, S, M> for Either<P1, P2>
+{
     type Output = P1::Output;
     #[inline]
     fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<P1::Output, I, S, M> {
@@ -124,90 +132,29 @@ impl<O: Clone, I: Input, C, S, M: Cb> Parser<I, C, S, M> for Pure<O> {
     }
 }
 
-#[derive(Clone, Copy)]
-pub enum Error<E> {
-    Unexpect(E),
-    Message(E),
-}
-
-/// A parser that never fails. You can choose between `Error::Unexpect`, which prints "Unexpected [token]", and `Error::Message`, which prints "[msg]".
+/// A parser that never fails. You can choose between [`error::unexpect`][`crate::error::unexpect`], which prints "Unexpected \[token\]", and [`error::message`][`crate::error::message`], which prints "\[msg\]".
+/// See also [`error::Builder`][`crate::error::Builder`]
 /// ```
 /// use chasa::*;
-/// assert_eq!(fail::<(),_>(prim::Error::Unexpect("chasa")).parse_easy("".chars()),Err("unexpected chasa at 0".to_string()));
+/// assert_eq!(fail::<_,()>(unexpect("chasa")).parse_easy(""),Err("unexpected chasa at 0".to_string()));
 /// ```
-pub struct Fail<O, E>(Error<E>, PhantomData<fn() -> O>);
-impl<O, E: Clone> Clone for Fail<O, E> {
-    #[inline]
-    fn clone(&self) -> Self {
-        Fail(self.0.clone(), PhantomData)
-    }
+pub struct Fail<O, M: Cb>(Eb<M>, PhantomData<fn() -> O>);
+#[inline]
+pub fn fail<M: Cb, O>(err: Eb<M>) -> Fail<O, M> {
+    Fail(err, PhantomData)
 }
-impl<O, E: Copy> Copy for Fail<O, E> {}
-pub fn fail<O, E: Display + 'static>(e: Error<E>) -> Fail<O, E> {
-    Fail(e, PhantomData)
-}
-impl<O, I: Input, C, S, M: Cb, E: Display + 'static> ParserOnce<I, C, S, M> for Fail<O, E> {
+impl<O, I: Input, C, S, M: Cb> ParserOnce<I, C, S, M> for Fail<O, M> {
     type Output = O;
     fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<O, I, S, M> {
-        Err(match self.0 {
-            Error::Unexpect(token) => Builder::unexpected(token),
-            Error::Message(msg) => Builder::message(msg),
-        }
-        .at::<I>(cont.ok.input.index(), cont.ok.input.pos(), None))
-    }
-}
-impl<O, I: Input, C, S, M: Cb, E: Display + Clone + 'static> Parser<I, C, S, M> for Fail<O, E> {
-    fn run(&self, cont: ICont<I, C, S, M>) -> IResult<O, I, S, M> {
-        Err(match &self.0 {
-            Error::Unexpect(token) => Builder::unexpected(token.clone()),
-            Error::Message(msg) => Builder::message(msg.clone()),
-        }
-        .at::<I>(cont.ok.input.index(), cont.ok.input.pos(), None))
-    }
-}
-
-/// An absolute failure parser, which delays error messages by passing a closure. You can choose between `Error::Unexpect` to show "Unexpected [token]", and `Error::Message` to show "[msg]".
-/// ```
-/// use chasa::*;
-/// assert_eq!(fail_with::<(),_,_>(prim::Error::Unexpect(||"chasa")).parse_easy("".chars()),Err("unexpected chasa at 0".to_string()));
-/// ```
-pub struct FailWith<O, F>(Error<F>, PhantomData<fn() -> O>);
-impl<O, F: Clone> Clone for FailWith<O, F> {
-    #[inline]
-    fn clone(&self) -> Self {
-        FailWith(self.0.clone(), PhantomData)
-    }
-}
-impl<O, F: Copy> Copy for FailWith<O, F> {}
-pub fn fail_with<O, E: Display, F: Fn() -> E + 'static>(f: Error<F>) -> FailWith<O, F> {
-    FailWith(f, PhantomData)
-}
-impl<O, I: Input, C, S, M: Cb, E: Display, F: Fn() -> E + 'static> ParserOnce<I, C, S, M> for FailWith<O, F> {
-    type Output = O;
-    fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<O, I, S, M> {
-        Err(match self.0 {
-            Error::Unexpect(token) => Builder::unexpected_with(token),
-            Error::Message(msg) => Builder::message_with(msg),
-        }
-        .at::<I>(cont.ok.input.index(), cont.ok.input.pos(), None))
-    }
-}
-impl<O, I: Input, C, S, M: Cb, E: Display, F: Fn() -> E + Clone + 'static> Parser<I, C, S, M> for FailWith<O, F> {
-    #[inline]
-    fn run(&self, cont: ICont<I, C, S, M>) -> IResult<O, I, S, M> {
-        Err(match &self.0 {
-            Error::Unexpect(token) => Builder::unexpected_with(token.clone()),
-            Error::Message(msg) => Builder::message_with(msg.clone()),
-        }
-        .at::<I>(cont.ok.input.index(), cont.ok.input.pos(), None))
+        Err(self.0.at::<I>(cont.ok.input.index(), cont.ok.input.pos(), None))
     }
 }
 
 /// The parser that matches the end of the input.
 /// ```
 /// use chasa::*;
-/// assert_eq!(eoi.parse_easy("".chars()), Ok(()));
-/// assert_eq!(eoi.parse_easy("a2".chars()), Err("unexpected a, expecting eoi at 0".to_string()))
+/// assert_eq!(eoi.parse_easy(""), Ok(()));
+/// assert_eq!(eoi.parse_easy("a2"), Err("unexpected a, expecting eoi at 0".to_string()))
 /// ```
 pub struct EoI<I, C, S, M>(PhantomData<fn() -> (I, C, S, M)>);
 impl<I, C, S, M> Clone for EoI<I, C, S, M> {
@@ -244,8 +191,8 @@ impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb> Parser<I, C, S, M> fo
 /// A parser that takes a single arbitrary character.
 /// ```
 /// use chasa::*;
-/// assert_eq!(any.parse_ok("12".chars()), Some('1'));
-/// assert_eq!(any.parse_ok("".chars()), None)
+/// assert_eq!(any.parse_ok("12"), Some('1'));
+/// assert_eq!(any.parse_ok(""), None)
 /// ```
 pub struct Any<I, C, S, M>(PhantomData<fn() -> (I, C, S, M)>);
 impl<I, C, S, M> Clone for Any<I, C, S, M> {
@@ -270,6 +217,7 @@ impl<I: Input, C, S, M: Cb> Parser<I, C, S, M> for Any<I, C, S, M> {
     #[inline]
     fn run(&self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
         let ICont { ok: IOk { mut input, err, state, .. }, drop, .. } = cont;
+
         if cont.ok.cutted {
             drop()
         }
@@ -288,8 +236,8 @@ impl<I: Input, C, S, M: Cb> Parser<I, C, S, M> for Any<I, C, S, M> {
 /// If the given character matches the input character, it is accepted.
 /// ```
 /// use chasa::*;
-/// assert_eq!(char('a').parse_easy("a2".chars()), Ok('a'));
-/// assert_eq!(char('1').parse_easy("a2".chars()), Err("unexpected a at 0..1".to_string()))
+/// assert_eq!(char('a').parse_easy("a2"), Ok('a'));
+/// assert_eq!(char('1').parse_easy("a2"), Err("unexpected a, expecting 1 at 0..1".to_string()))
 /// ```
 pub struct Char<Item, I, C, S, M>(Item, PhantomData<fn() -> (I, C, S, M)>);
 impl<Item: Clone, I, C, S, M> Clone for Char<Item, I, C, S, M> {
@@ -302,32 +250,26 @@ impl<Item: Copy, I, C, S, M> Copy for Char<Item, I, C, S, M> {}
 pub fn char<Item, I, C, S, M>(char: Item) -> Char<Item, I, C, S, M> {
     Char(char, PhantomData)
 }
-impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::Item>> ParserOnce<I, C, S, M> for Char<Item, I, C, S, M> {
+impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::Item> + Display> ParserOnce<I, C, S, M>
+    for Char<Item, I, C, S, M>
+{
     type Output = I::Item;
     #[inline]
     fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
         self.run(cont)
     }
 }
-impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::Item>> Parser<I, C, S, M> for Char<Item, I, C, S, M> {
+impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::Item> + Display> Parser<I, C, S, M>
+    for Char<Item, I, C, S, M>
+{
     #[inline]
     fn run(&self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
         let ICont { ok: IOk { mut input, err, state, .. }, drop, .. } = cont;
-        if cont.ok.cutted {
-            drop()
-        }
-        let (index, pos) = (input.index(), input.pos());
-        match input.next() {
-            None => Err(Eb::unexpected_eoi().at::<I>(index, pos, None).or_merge(err)),
-            Some(Err(e)) => Err(Eb::message(e).at::<I>(index, pos, None).or_merge(err)),
-            Some(Ok(item)) => {
-                if &self.0 == &item {
-                    drop();
-                    Ok((item, IOk { input, state, err, cutted: true }))
-                } else {
-                    Err(Eb::unexpected(item).at::<I>(input.index(), pos, Some(input.pos())).or_merge(err))
-                }
-            },
+
+        match run_satisfy(&mut input, drop, cont.ok.cutted, |item| if &self.0 == &item { Ok(item) } else { Err(item) })
+        {
+            Err(e) => Err(e.label(format!("{}", self.0)).or_merge(err)),
+            Ok(item) => Ok((item, IOk { input, state, err, cutted: true })),
         }
     }
 }
@@ -335,8 +277,8 @@ impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::It
 /// It takes one character from the input, compares it with the given iterator, and accepts it if any of the characters match.
 /// ```
 /// use chasa::*;
-/// assert_eq!(one_of("abc".chars()).parse_ok("a2".chars()), Some('a'));
-/// assert_eq!(one_of("def".chars()).parse_ok("a2".chars()), None)
+/// assert_eq!(one_of("abc").parse_ok("a2"), Some('a'));
+/// assert_eq!(one_of("def").parse_ok("a2"), None)
 /// ```
 pub struct OneOf<Iter, Item, I, C, S, M>(Iter, PhantomData<fn() -> (Item, I, C, S, M)>);
 impl<Iter: Clone, Item, I, C, S, M> Clone for OneOf<Iter, Item, I, C, S, M> {
@@ -347,10 +289,21 @@ impl<Iter: Clone, Item, I, C, S, M> Clone for OneOf<Iter, Item, I, C, S, M> {
 }
 impl<Iter: Copy, Item, I, C, S, M> Copy for OneOf<Iter, Item, I, C, S, M> {}
 #[inline]
-pub fn one_of<Iter: IntoIterator<Item = Item>, Item, I, C, S, M>(iter: Iter) -> OneOf<Iter, Item, I, C, S, M> {
-    OneOf(iter, PhantomData)
+pub fn one_of<Iter, Item, I, C, S, M>(chars: Iter) -> OneOf<Iter, Item, I, C, S, M> {
+    OneOf(chars, PhantomData)
 }
-impl<I: Input, C, S, M: Cb, Item: PartialEq<I::Item>, Iter: IntoIterator<Item = Item>> ParserOnce<I, C, S, M> for OneOf<Iter, Item, I, C, S, M>
+impl<I: Input, C, S, M: Cb, Item, Iter> Parser<I, C, S, M> for OneOf<Iter, Item, I, C, S, M>
+where
+    I::Item: Display + 'static,
+    Self: ParserOnce<I, C, S, M, Output = I::Item> + Clone,
+{
+    #[inline]
+    fn run(&self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
+        self.clone().run_once(cont)
+    }
+}
+impl<I: Input, C, S, M: Cb, Item: PartialEq<I::Item>, Iter: IntoChars<Item = Item>> ParserOnce<I, C, S, M>
+    for OneOf<Iter, Item, I, C, S, M>
 where
     I::Item: Display + 'static,
 {
@@ -358,40 +311,68 @@ where
     #[inline]
     fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
         let ICont { ok: IOk { mut input, err, state, .. }, drop, .. } = cont;
-        if cont.ok.cutted {
-            drop()
-        }
-        let (index, pos) = (input.index(), input.pos());
-        match input.next() {
-            None => Err(Eb::unexpected_eoi().at::<I>(index, pos, None).or_merge(err)),
-            Some(Err(e)) => Err(Eb::message(e).at::<I>(index, pos, None).or_merge(err)),
-            Some(Ok(item)) => {
-                for compared in self.0.into_iter() {
-                    if &compared == &item {
-                        drop();
-                        return Ok((item, IOk { input, state, err, cutted: true }));
-                    }
+        match run_satisfy(&mut input, drop, cont.ok.cutted, |item| {
+            for compared in self.0.into_chars() {
+                if &compared == &item {
+                    return Ok(item);
                 }
-                Err(Eb::unexpected(item).at::<I>(input.index(), pos, Some(input.pos())).or_merge(err))
-            },
+            }
+            Err(item)
+        }) {
+            Ok(item) => Ok((item, IOk { input, state, err, cutted: true })),
+            Err(e) => Err(e.or_merge(err)),
         }
     }
 }
-impl<I: Input, C, S, M: Cb, Item: PartialEq<I::Item>, Iter: IntoIterator<Item = Item> + Clone> Parser<I, C, S, M> for OneOf<Iter, Item, I, C, S, M>
-where
-    I::Item: Display + 'static,
-{
-    #[inline]
-    fn run(&self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
-        self.clone().run_once(cont)
+macro_rules! one_of_range {
+    ($item:ident, $t:ty) =>{
+        impl<I: Input, C, S, M: Cb, $item: PartialOrd<I::Item>> ParserOnce<I, C, S, M>
+            for OneOf<$t, Item, I, C, S, M>
+        where
+            I::Item: PartialOrd<Item> + Display + 'static,
+        {
+            type Output = I::Item;
+            #[inline]
+            fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
+                let ICont { ok: IOk { mut input, err, state, .. }, drop, .. } = cont;
+                match run_satisfy(
+                    &mut input,
+                    drop,
+                    cont.ok.cutted,
+                    |item| {
+                        if self.0.contains(&item) {
+                            Ok(item)
+                        } else {
+                            Err(item)
+                        }
+                    },
+                ) {
+                    Ok(item) => Ok((item, IOk { input, state, err, cutted: true })),
+                    Err(e) => Err(e.or_merge(err)),
+                }
+            }
+        }
+    };
+    ($iter:ident, $t:ty, $($ts:ty),+) => {
+        one_of_range!($iter,$t);
+        one_of_range!($iter,$($ts),+);
     }
 }
+one_of_range!(
+    Item,
+    std::ops::Range<Item>,
+    std::ops::RangeFrom<Item>,
+    std::ops::RangeTo<Item>,
+    std::ops::RangeFull,
+    std::ops::RangeInclusive<Item>,
+    std::ops::RangeToInclusive<Item>
+);
 
 /// It takes one character from the input, compares it with the given iterator, and only accepts if none of the characters match.
 /// ```
 /// use chasa::*;
-/// assert_eq!(none_of("abc".chars()).parse_ok("a2".chars()), None);
-/// assert_eq!(none_of("def".chars()).parse_ok("a2".chars()), Some('a'))
+/// assert_eq!(none_of("abc").parse_ok("a2"), None);
+/// assert_eq!(none_of("def").parse_ok("a2"), Some('a'))
 /// ```
 pub struct NoneOf<Iter, Item, I, C, S, M>(Iter, PhantomData<fn() -> (Item, I, C, S, M)>);
 impl<Iter: Clone, Item, I, C, S, M> Clone for NoneOf<Iter, Item, I, C, S, M> {
@@ -402,11 +383,21 @@ impl<Iter: Clone, Item, I, C, S, M> Clone for NoneOf<Iter, Item, I, C, S, M> {
 }
 impl<Iter: Copy, Item, I, C, S, M> Copy for NoneOf<Iter, Item, I, C, S, M> {}
 #[inline]
-pub fn none_of<Iter: IntoIterator<Item = Item>, Item, I, C, S, M>(iter: Iter) -> NoneOf<Iter, Item, I, C, S, M> {
-    NoneOf(iter, PhantomData)
+pub fn none_of<Iter: IntoChars<Item = Item>, Item, I, C, S, M>(chars: Iter) -> NoneOf<Iter, Item, I, C, S, M> {
+    NoneOf(chars, PhantomData)
 }
-impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::Item>, Iter: IntoIterator<Item = Item>> ParserOnce<I, C, S, M>
+impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item, Iter> Parser<I, C, S, M>
     for NoneOf<Iter, Item, I, C, S, M>
+where
+    Self: ParserOnce<I, C, S, M, Output = I::Item> + Clone,
+{
+    #[inline]
+    fn run(&self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
+        self.clone().run_once(cont)
+    }
+}
+impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::Item>, Iter: IntoChars<Item = Item>>
+    ParserOnce<I, C, S, M> for NoneOf<Iter, Item, I, C, S, M>
 {
     type Output = I::Item;
     #[inline]
@@ -420,7 +411,7 @@ impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::It
             None => Err(Eb::unexpected_eoi().at::<I>(index, pos, None).or_merge(err)),
             Some(Err(e)) => Err(Eb::message(e).at::<I>(index, pos, None).or_merge(err)),
             Some(Ok(item)) => {
-                for compared in self.0.into_iter() {
+                for compared in self.0.into_chars() {
                     if &compared == &item {
                         return Err(Eb::unexpected(item).at::<I>(input.index(), pos, Some(input.pos())).or_merge(err));
                     }
@@ -430,43 +421,78 @@ impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::It
         }
     }
 }
-impl<I: Input<Item = impl Display + 'static>, C, S, M: Cb, Item: PartialEq<I::Item>, Iter: IntoIterator<Item = Item> + Clone> Parser<I, C, S, M>
-    for NoneOf<Iter, Item, I, C, S, M>
-{
-    #[inline]
-    fn run(&self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
-        self.clone().run_once(cont)
+macro_rules! none_of_range {
+    ($item:ident,$t:ty) => {
+        impl<I: Input<Item = impl PartialOrd<Item> + Display + 'static>, C, S, M: Cb, $item: PartialOrd<I::Item>> ParserOnce<I, C, S, M>
+            for NoneOf<$t, Item, I, C, S, M>
+        {
+            type Output = I::Item;
+            #[inline]
+            fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<I::Item, I, S, M> {
+                let ICont { ok: IOk { mut input, err, state, .. }, drop, .. } = cont;
+                match run_satisfy(&mut input, drop, cont.ok.cutted, |item| {
+                    if !self.0.contains(&item) {
+                        Ok(item)
+                    } else {
+                        Err(item)
+                    }
+                }) {
+                    Ok(item) => Ok((item, IOk { input, state, err, cutted: true })),
+                    Err(e) => Err(e.or_merge(err)),
+                }
+            }
+        }
+    };
+    ($item:ident, $t:ty, $($ts:ty),+) => {
+        none_of_range!($item,$t);
+        none_of_range!($item,$($ts),+);
     }
 }
+none_of_range!(
+    Item,
+    std::ops::Range<Item>,
+    std::ops::RangeFrom<Item>,
+    std::ops::RangeTo<Item>,
+    std::ops::RangeFull,
+    std::ops::RangeInclusive<Item>,
+    std::ops::RangeToInclusive<Item>
+);
 
 /// A parser that compares character iterators and input as they are consumed together, and accepts them if they all match.
+/// The return value is empty, as most of the time it should be intended to be a token.
+/// If you want the whole string, use `.to([string])`(see [`Value`][`crate::combi::Value`]) or `.get_str()`(see [`GetString`][`crate::combi::GetString`]).
 /// ```
 /// use chasa::*;
-/// assert_eq!(string("a".chars(), 1).parse_ok("a2".chars()), Some(1));
-/// assert_eq!(string("a2".chars(), 1).parse_ok("a2".chars()), Some(1));
-/// assert_eq!(string("a23".chars(), 1).parse_ok("a2".chars()), None);
-/// assert_eq!(string("a3".chars(), 1).or(string("a".chars(), 2)).parse_ok("a2".chars()), Some(2));
+/// assert_eq!(str("a").parse_ok("a2"), Some(()));
+/// assert_eq!(str("a2").parse_ok("a2"), Some(()));
+/// assert_eq!(str("a23").parse_ok("a2"), None);
+/// assert_eq!(str("a3").to(1).or(str("a").to(2)).parse_ok("a2"), Some(2));
 /// ```
-pub struct String<Iter, O, I, C, S, M>(Iter, O, PhantomData<fn() -> (I, C, S, M)>);
-impl<Iter: Clone, O: Clone, I, C, S, M> Clone for String<Iter, O, I, C, S, M> {
+pub struct String<Iter, I, C, S, M>(Iter, PhantomData<fn() -> (I, C, S, M)>);
+impl<Iter: Clone, I, C, S, M> Clone for String<Iter, I, C, S, M> {
     #[inline]
     fn clone(&self) -> Self {
-        String(self.0.clone(), self.1.clone(), PhantomData)
+        String(self.0.clone(), PhantomData)
     }
 }
-impl<Iter: Copy, O: Copy, I, C, S, M> Copy for String<Iter, O, I, C, S, M> {}
-pub fn string<Iter: IntoIterator<Item = I::Item>, O, I: Input, C, S, M: Cb>(iter: Iter, value: O) -> String<Iter, O, I, C, S, M> {
-    String(iter, value, PhantomData)
+impl<Iter: Copy, I, C, S, M> Copy for String<Iter, I, C, S, M> {}
+pub fn str<Iter: IntoChars<Item = I::Item>, I: Input, C, S, M: Cb>(iter: Iter) -> String<Iter, I, C, S, M> {
+    String(iter, PhantomData)
 }
-impl<O, I: Input<Item = impl Display + 'static>, C, S, M: Cb, Iter: IntoIterator<Item = impl PartialEq<I::Item> + Display + 'static>>
-    ParserOnce<I, C, S, M> for String<Iter, O, I, C, S, M>
+impl<
+        I: Input<Item = impl Display + 'static>,
+        C,
+        S,
+        M: Cb,
+        Iter: IntoChars<Item = impl PartialEq<I::Item> + Display + 'static>,
+    > ParserOnce<I, C, S, M> for String<Iter, I, C, S, M>
 {
-    type Output = O;
+    type Output = ();
     #[inline]
-    fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<O, I, S, M> {
+    fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<(), I, S, M> {
         let ICont { ok: IOk { mut input, err, state, .. }, drop, .. } = cont;
         let mut is_first = true;
-        for str_c in self.0 {
+        for str_c in self.0.into_chars() {
             if is_first {
                 if cont.ok.cutted {
                     drop()
@@ -479,25 +505,27 @@ impl<O, I: Input<Item = impl Display + 'static>, C, S, M: Cb, Iter: IntoIterator
                 Some(Err(e)) => return Err(Eb::message(e).at::<I>(index, pos, None).or_merge(err)),
                 Some(Ok(input_c)) => {
                     if str_c != input_c {
-                        return Err(Eb::unexpected(input_c).label(str_c).at::<I>(input.index(), pos, Some(input.pos())).or_merge(err));
+                        return Err(Eb::unexpected(input_c)
+                            .label(str_c)
+                            .at::<I>(input.index(), pos, Some(input.pos()))
+                            .or_merge(err));
                     }
                 },
             }
         }
-        Ok((self.1, IOk { input, err: if is_first { err } else { None }, state, cutted: !is_first }))
+        Ok(((), IOk { input, err: if is_first { err } else { None }, state, cutted: !is_first }))
     }
 }
 impl<
-        O: Clone,
         I: Input<Item = impl Display + 'static>,
         C,
         S,
         M: Cb,
-        Iter: IntoIterator<Item = impl PartialEq<I::Item> + Display + 'static> + Clone,
-    > Parser<I, C, S, M> for String<Iter, O, I, C, S, M>
+        Iter: IntoChars<Item = impl PartialEq<I::Item> + Display + 'static> + Clone,
+    > Parser<I, C, S, M> for String<Iter, I, C, S, M>
 {
     #[inline]
-    fn run(&self, cont: ICont<I, C, S, M>) -> IResult<O, I, S, M> {
+    fn run(&self, cont: ICont<I, C, S, M>) -> IResult<(), I, S, M> {
         self.clone().run(cont)
     }
 }
@@ -505,7 +533,11 @@ impl<
 /// A parser that takes a single character satisfying a condition.
 pub struct Satisfy<F, I, C, S, M>(F, PhantomData<fn() -> (I, C, S, M)>);
 #[inline]
-pub fn satisfy<F, I, C, S, M>(f: F) -> Satisfy<F, I, C, S, M> {
+pub fn satisfy<I: Input, F: Fn(&I::Item) -> bool, C, S, M>(f: F) -> Satisfy<F, I, C, S, M> {
+    Satisfy(f, PhantomData)
+}
+#[inline]
+pub fn satisfy_once<I: Input, F: FnOnce(&I::Item) -> bool, C, S, M>(f: F) -> Satisfy<F, I, C, S, M> {
     Satisfy(f, PhantomData)
 }
 impl<F: Clone, I, C, S, M> Clone for Satisfy<F, I, C, S, M> {
@@ -569,7 +601,11 @@ where
 
 pub struct SatisfyMap<F, I, C, S, M>(F, PhantomData<fn() -> (I, C, S, M)>);
 #[inline]
-pub fn satisfy_map<F, I, C, S, M>(f: F) -> SatisfyMap<F, I, C, S, M> {
+pub fn satisfy_map<F: Fn(&I::Item) -> Option<O>, O, I: Input, C, S, M>(f: F) -> SatisfyMap<F, I, C, S, M> {
+    SatisfyMap(f, PhantomData)
+}
+#[inline]
+pub fn satisfy_map_once<F: FnOnce(&I::Item) -> Option<O>, O, I: Input, C, S, M>(f: F) -> SatisfyMap<F, I, C, S, M> {
     SatisfyMap(f, PhantomData)
 }
 impl<F: Clone, I, C, S, M> Clone for SatisfyMap<F, I, C, S, M> {
@@ -771,7 +807,9 @@ pub fn local_state<S, SLocal, P>(state: SLocal, parser: P) -> LocalState<S, SLoc
 pub fn no_state<S, P>(parser: P) -> LocalState<S, (), P> {
     LocalState((), parser, PhantomData)
 }
-impl<I: Input, C, S, SLocal, M: Cb, P: ParserOnce<I, C, SLocal, M>> ParserOnce<I, C, S, M> for LocalState<S, SLocal, P> {
+impl<I: Input, C, S, SLocal, M: Cb, P: ParserOnce<I, C, SLocal, M>> ParserOnce<I, C, S, M>
+    for LocalState<S, SLocal, P>
+{
     type Output = P::Output;
     #[inline]
     fn run_once(self, cont: ICont<I, C, S, M>) -> IResult<P::Output, I, S, M> {
